from abc import ABC, abstractmethod
from hashlib import sha256
import os
from pathlib import Path
import random
from time import sleep, time
from uuid import uuid4
from morecontext import envset
from fscacher import PersistentCache


class BaseCacheBenchmark(ABC):
    param_names = ["mode"]
    params = [["populate", "hit", "ignore"]]

    @abstractmethod
    def init_path(self, *args):
        # Must return the path created
        ...

    @staticmethod
    @abstractmethod
    def init_func(cache):
        # Must return the function
        ...

    def init_cache(self, ignore: bool = False):
        with envset("FSCACHER_CACHE", "ignore" if ignore else ""):
            self.cache = PersistentCache(path=str(uuid4()))
        self.func = self.init_func(self.cache)

    def setup(self, mode, *args):
        self.path = self.init_path(mode, *args)
        if mode == "hit":
            self.init_cache()
            self.func(self.path)
        elif mode == "ignore":
            self.init_cache(ignore=True)

    def time_cache(self, mode, *_args):
        if mode == "populate":
            self.init_cache()
        self.func(self.path)


class TimeFile(BaseCacheBenchmark):
    FILE_SIZE = 1024

    def init_path(self, *_args):
        with open("foo.dat", "wb") as fp:
            fp.write(bytes(random.choices(range(256), k=self.FILE_SIZE)))
        return "foo.dat"

    @staticmethod
    def init_func(cache):
        @cache.memoize_path
        def hashfile(path):
            # "emulate" slow invocation so significant raise in benchmark
            # consumed time would mean that we invoked it instead
            # of using cached value
            sleep(0.01)
            with open(path, "rb") as fp:
                return sha256(fp.read()).hexdigest()

        return hashfile


class BaseDirectoryBenchmark(BaseCacheBenchmark):
    param_names = BaseCacheBenchmark.param_names + ["tmpdir"]
    params = BaseCacheBenchmark.params + [
        os.environ.get("FSCACHER_BENCH_TMPDIRS", ".").split(":")
    ]

    @staticmethod
    @abstractmethod
    def get_layout():
        ...

    def init_path(self, _mode, tmpdir):
        dirpath = Path(tmpdir, str(uuid4()))
        dirpath.mkdir(parents=True)
        base_time = time()
        dirs = [dirpath]
        layout = self.get_layout()
        for i, width in enumerate(layout):
            if i < len(layout) - 1:
                dirs2 = []
                for d in dirs:
                    for x in range(width):
                        d2 = d / f"d{x}"
                        d2.mkdir()
                        dirs2.append(d2)
                dirs = dirs2
            else:
                for j, d in enumerate(dirs):
                    for x in range(width):
                        f = d / f"f{x}.dat"
                        f.write_bytes(b"\0" * random.randint(1, 1024))
                        t = base_time - x - j * width
                        os.utime(f, (t, t))
        return dirpath

    @staticmethod
    def init_func(cache):
        @cache.memoize_path
        def dirsize(path):
            total_size = 0
            with os.scandir(path) as entries:
                for e in entries:
                    if e.is_dir():
                        total_size += dirsize(e.path)
                    else:
                        total_size += e.stat().st_size
            return total_size

        return dirsize


class TimeFlatDirectory(BaseDirectoryBenchmark):
    @staticmethod
    def get_layout():
        return (100,)


class TimeDeepDirectory(BaseDirectoryBenchmark):
    @staticmethod
    def get_layout():
        return (3, 3, 3, 3)



from collections import deque, namedtuple
from functools import partial, wraps
from hashlib import md5
from inspect import Parameter, signature
import logging
import os
import os.path as op
import shutil
import time
import joblib
from platformdirs import PlatformDirs

lgr = logging.getLogger(__name__)


class PersistentCache(object):
    """Persistent cache providing @memoize and @memoize_path decorators"""

    _min_dtime = 0.01  # min difference between now and mtime to consider
    # for caching

    _cache_var_values = (None, "", "clear", "ignore")

    def __init__(self, name=None, *, path=None, tokens=None, envvar=None):
        """
        Parameters
        ----------
        name: str, optional
         Basename for the directory in which to store the cache.  Mutually
         exclusive with `path`.
        path: str or pathlib.Path, optional
         Directory path at which to store the cache.  If not specified, the
         cache is stored in `USER_CACHE/fscacher/NAME`, where NAME is the value
         of the `name` parameter (default: "cache").  Mutually exclusive with
         `name`.
        tokens: list of objects, optional
         To add to the fingerprint of @memoize_path (regular @memoize ATM does
         not use it).  Could be e.g. versions of relevant/used
         python modules (pynwb, etc)
        envvar: str, optional
         Name of the environment variable to query for cache settings; if not
         set, `FSCACHER_CACHE` is used
        """
        if path is None:
            dirs = PlatformDirs("fscacher")
            path = op.join(dirs.user_cache_dir, (name or "cache"))
        elif name is not None:
            raise ValueError("'name' and 'path' are mutually exclusive")
        self._memory = joblib.Memory(path, verbose=0)
        cntrl_value = None
        if envvar is not None:
            cntrl_var = envvar
            cntrl_value = os.environ.get(cntrl_var)
        if cntrl_value is None:
            cntrl_var = "FSCACHER_CACHE"
            cntrl_value = os.environ.get(cntrl_var)
        if cntrl_value not in self._cache_var_values:
            lgr.warning(
                f"{cntrl_var}={cntrl_value} is not understood and thus ignored."
                f" Known values are {self._cache_var_values}"
            )
        if cntrl_value == "clear":
            self.clear()
        self._ignore_cache = cntrl_value == "ignore"
        self._tokens = tokens

    def clear(self):
        try:
            self._memory.clear(warn=False)
        except Exception as exc:
            lgr.debug("joblib failed to clear its cache: %s", exc)

        # and completely destroy the directory
        try:
            if op.exists(self._memory.location):
                shutil.rmtree(self._memory.location)
        except Exception as exc:
            lgr.warning(f"Failed to clear out the cache directory: {exc}")

    def memoize(self, f=None, *, exclude_kwargs=None):
        if f is None:
            return partial(self.memoize, exclude_kwargs=exclude_kwargs)
        if self._ignore_cache:
            return f
        return self._memory.cache(f, ignore=exclude_kwargs)

    def memoize_path(self, f=None, *, exclude_kwargs=None):
        if f is None:
            return partial(self.memoize_path, exclude_kwargs=exclude_kwargs)
        if self._ignore_cache:
            return f

        # we need to actually decorate a function
        fingerprint_kwarg = "_cache_fingerprint"

        @wraps(f)  # important, so .memoize correctly caches different `f`
        def fingerprinted(path, *args, **kwargs):
            _ = kwargs.pop(fingerprint_kwarg)  # discard
            lgr.debug("Running original %s on %r", f, path)
            return f(path, *args, **kwargs)

        # We need to add the fingerprint_kwarg to fingerprinted's signature so
        # that joblib doesn't complain:
        sig = signature(fingerprinted)
        fp_kwarg_param = Parameter(
            fingerprint_kwarg, Parameter.KEYWORD_ONLY, default=None
        )
        sig2 = sig.replace(
            parameters=tuple(sig.parameters.values()) + (fp_kwarg_param,)
        )
        fingerprinted.__signature__ = sig2

        path_arg = next(iter(sig.parameters.keys()))

        # we need to ignore 'path' since we would like to dereference if symlink
        # but then expect joblib's caching work on both original and dereferenced
        # So we will add dereferenced path into fingerprint_kwarg
        fingerprinted = self.memoize(
            fingerprinted,
            exclude_kwargs=[path_arg]
            + (list(exclude_kwargs) if exclude_kwargs is not None else []),
        )

        @wraps(f)
        def fingerprinter(*args, **kwargs):
            # we need to dereference symlinks and use that path in the function
            # call signature
            bound = sig.bind(*args, **kwargs)
            bound.apply_defaults()
            path_orig = bound.arguments[path_arg]
            try:
                path = op.realpath(path_orig)
            except TypeError:
                lgr.debug(
                    "Calling %s directly since argument is not a path-like object", f
                )
                return f(*args, **kwargs)
            if path != path_orig:
                lgr.log(5, "Dereferenced %r into %r", path_orig, path)
            if op.isdir(path):
                fprint = self._get_dir_fingerprint(path)
            else:
                fprint = self._get_file_fingerprint(path)
            if fprint is None:
                lgr.debug("Calling %s directly since no fingerprint for %r", f, path)
                # just call the function -- we have no fingerprint,
                # probably does not exist or permissions are wrong
                ret = f(*args, **kwargs)
            # We should still pass through if file was modified just now,
            # since that could mask out quick modifications.
            # Target use cases will not be like that.
            elif fprint.modified_in_window(self._min_dtime):
                lgr.debug("Calling %s directly since too short for %r", f, path)
                ret = f(*args, **kwargs)
            else:
                lgr.debug("Calling memoized version of %s for %s", f, path)
                # If there is a fingerprint -- inject it into the signature
                kwargs_ = kwargs.copy()
                kwargs_[fingerprint_kwarg] = (
                    (path,)
                    + fprint.to_tuple()
                    + (tuple(self._tokens) if self._tokens else ())
                )
                ret = fingerprinted(*args, **kwargs_)
            lgr.log(1, "Returning value %r", ret)
            return ret

        # and we memoize actually that function
        return fingerprinter

    @staticmethod
    def _get_file_fingerprint(path):
        """Simplistic generic file fingerprinting based on ctime, mtime, and size"""
        try:
            # we can't take everything, since atime can change, etc.
            # So let's take some
            s = os.stat(path, follow_symlinks=True)
            fprint = FileFingerprint.from_stat(s)
            lgr.log(5, "Fingerprint for %s: %s", path, fprint)
            return fprint
        except Exception as exc:
            lgr.debug(f"Cannot fingerprint {path}: {exc}")

    @staticmethod
    def _get_dir_fingerprint(path):
        fprint = DirFingerprint()
        dirqueue = deque([path])
        try:
            while dirqueue:
                d = dirqueue.popleft()
                with os.scandir(d) as entries:
                    for e in entries:
                        if e.is_dir(follow_symlinks=True):
                            dirqueue.append(e.path)
                        else:
                            s = e.stat(follow_symlinks=True)
                            fprint.add_file(e.path, FileFingerprint.from_stat(s))
        except Exception as exc:
            lgr.debug(f"Cannot fingerprint {path}: {exc}")
            return None
        else:
            return fprint


class FileFingerprint(namedtuple("FileFingerprint", "mtime_ns ctime_ns size inode")):
    @classmethod
    def from_stat(cls, s):
        return cls(s.st_mtime_ns, s.st_ctime_ns, s.st_size, s.st_ino)

    def modified_in_window(self, min_dtime):
        return abs(elapsed_since(self.mtime_ns * 1e-9)) < min_dtime

    def to_tuple(self):
        return tuple(self)


class DirFingerprint:
    def __init__(self):
        self.last_modified = None
        self.hash = None

    def add_file(self, path, fprint: FileFingerprint):
        fprint_hash = md5(
            ascii((str(path), fprint.to_tuple())).encode("us-ascii")
        ).digest()
        if self.hash is None:
            self.hash = fprint_hash
            self.last_modified = fprint.mtime_ns
        else:
            self.hash = xor_bytes(self.hash, fprint_hash)
            if self.last_modified < fprint.mtime_ns:
                self.last_modified = fprint.mtime_ns

    def modified_in_window(self, min_dtime):
        if self.last_modified is None:
            return False
        else:
            return abs(elapsed_since(self.last_modified * 1e-9)) < min_dtime

    def to_tuple(self):
        if self.hash is None:
            return (None,)
        else:
            return (self.hash.hex(),)


def xor_bytes(b1: bytes, b2: bytes) -> bytes:
    length = max(len(b1), len(b2))
    # force 'little' byte order to match our assumptions on how to
    # treat bytes of different length.
    i1 = int.from_bytes(b1, "little")
    i2 = int.from_bytes(b2, "little")
    return (i1 ^ i2).to_bytes(length, "little")


def elapsed_since(t: float) -> float:
    t_now = time.time()
    dt = t_now - t
    if dt < 0:
        lgr.debug(
            "Time is in the future: %f; now: %f; dt=%g",
            t, t_now, dt
        )
    return dt



from itertools import zip_longest
import pytest
from ..cache import xor_bytes


@pytest.mark.parametrize(
    "b1,b2,r",
    [
        (b"\x12", b"\x34", b"\x26"),
        (b"\0\x12", b"\0\x34", b"\0\x26"),
        (b"\x12\0", b"\x34\0", b"\x26\0"),
        (b"\x12\xAB", b"\x34", b"\x26\xAB"),
        (b"\x12\xAB", b"\x34\xCD", b"\x26\x66"),
    ],
)
def test_xor_bytes(b1: bytes, b2: bytes, r: bytes) -> None:
    # test the target value assumption
    assert r == bytes(x ^ y for x, y in zip_longest(b1, b2, fillvalue=0))
    assert xor_bytes(b1, b2) == r



from dataclasses import dataclass
import logging
import os
import os.path as op
from pathlib import Path
import platform
import shutil
import subprocess
import sys
import time
import pytest
from .. import PersistentCache
from ..cache import DirFingerprint, FileFingerprint

platform_system = platform.system().lower()
on_windows = platform_system == "windows"
on_pypy = platform.python_implementation().lower() == "pypy"

lgr = logging.getLogger(__name__)


@pytest.fixture(autouse=True)
def capture_all_logs(caplog):
    caplog.set_level(1, logger="fscacher")


@pytest.fixture(scope="function")
def cache(tmp_path_factory):
    return PersistentCache(path=tmp_path_factory.mktemp("cache"))


@pytest.fixture(scope="function")
def cache_tokens(tmp_path_factory):
    return PersistentCache(path=tmp_path_factory.mktemp("cache"), tokens=["0.0.1", 1])


def test_memoize(cache):
    # Simplest testing to start with, not relying on persisting across
    # independent processes
    _comp = []

    @cache.memoize
    def f1(flag=False):
        if flag:
            raise ValueError("Got flag")
        if _comp:
            raise RuntimeError("Must not be recomputed")
        _comp.append(1)
        return 1

    assert f1() == 1
    assert f1() == 1

    # Now with some args
    _comp = []

    @cache.memoize
    def f2(*args):
        if args in _comp:
            raise RuntimeError("Must not be recomputed")
        _comp.append(args)
        return sum(args)

    assert f2(1) == 1
    assert f2(1) == 1
    assert f2(1, 2) == 3
    assert f2(1, 2) == 3
    assert _comp == [(1,), (1, 2)]


def test_memoize_multiple(cache):
    # Make sure that with the same cache can cover multiple functions
    @cache.memoize
    def f1():
        return 1

    @cache.memoize
    def f2():
        return 2

    @cache.memoize
    def f3():  # nesting call into f2
        return f2() + 1

    for _ in range(3):
        assert f1() == 1
        assert f2() == 2
        assert f3() == 3


def test_memoize_path(cache, tmp_path):
    calls = []

    @cache.memoize_path
    def memoread(path, arg, kwarg=None):
        calls.append([path, arg, kwarg])
        with open(path) as f:
            return f.read()

    def check_new_memoread(arg, content, expect_new=False):
        ncalls = len(calls)
        assert memoread(path, arg) == content
        assert len(calls) == ncalls + 1
        assert memoread(path, arg) == content
        assert len(calls) == ncalls + 1 + int(expect_new)

    fname = "file.dat"
    path = str(tmp_path / fname)

    with pytest.raises(IOError):
        memoread(path, 0)
    # and again
    with pytest.raises(IOError):
        memoread(path, 0)
    assert len(calls) == 2

    with open(path, "w") as f:
        f.write("content")

    t0 = time.time()
    try:
        # unless this computer is too slow -- there should be less than
        # cache._min_dtime between our creating the file and testing,
        # so we would force a direct read:
        check_new_memoread(0, "content", True)
    except AssertionError:  # pragma: no cover
        # if computer is indeed slow (happens on shared CIs) we might fail
        # because distance is too short
        if time.time() - t0 < cache._min_dtime:
            raise  # if we were quick but still failed -- legit
    assert calls[-1] == [path, 0, None]

    # but if we sleep - should memoize
    time.sleep(cache._min_dtime * 1.1)
    check_new_memoread(1, "content")

    # and if we modify the file -- a new read
    time.sleep(cache._min_dtime * 1.1)
    with open(path, "w") as f:
        f.write("Content")
    ncalls = len(calls)
    assert memoread(path, 1) == "Content"
    assert len(calls) == ncalls + 1

    time.sleep(cache._min_dtime * 1.1)
    check_new_memoread(0, "Content")

    # Check that symlinks should be dereferenced
    if not on_windows or (
        sys.version_info[:2] >= (3, 8) and not (on_windows and on_pypy)
    ):
        # realpath doesn't work right on Windows on pre-3.8 Python, and PyPy on
        # Windows doesn't support symlinks at all, so skip the test then.
        symlink1 = str(tmp_path / (fname + ".link1"))
        try:
            os.symlink(fname, symlink1)
        except OSError:
            pass
        if op.islink(symlink1):  # hopefully would just skip Windows if not supported
            ncalls = len(calls)
            assert memoread(symlink1, 0) == "Content"
            assert len(calls) == ncalls  # no new call

    # and if we "clear", would it still work?
    cache.clear()
    check_new_memoread(1, "Content")


@pytest.mark.flaky(reruns=5, condition=on_windows)
def test_memoize_path_dir(cache, tmp_path):
    calls = []

    @cache.memoize_path
    def memoread(path, arg, kwarg=None):
        calls.append([path, arg, kwarg])
        total_size = 0
        with os.scandir(path) as entries:
            for e in entries:
                if e.is_file():
                    total_size += e.stat().st_size
        return total_size

    def check_new_memoread(arg, content, expect_new=False):
        ncalls = len(calls)
        assert memoread(path, arg) == content
        assert len(calls) == ncalls + 1
        assert memoread(path, arg) == content
        assert len(calls) == ncalls + 1 + int(expect_new)

    fname = "foo"
    path = tmp_path / fname

    with pytest.raises(IOError):
        memoread(path, 0)
    # and again
    with pytest.raises(IOError):
        memoread(path, 0)
    assert len(calls) == 2

    path.mkdir()
    (path / "a.txt").write_text("Alpha")
    (path / "b.txt").write_text("Beta")

    t0 = time.time()
    try:
        # unless this computer is too slow -- there should be less than
        # cache._min_dtime between our creating the file and testing,
        # so we would force a direct read:
        check_new_memoread(0, 9, True)
    except AssertionError:  # pragma: no cover
        # if computer is indeed slow (happens on shared CIs) we might fail
        # because distance is too short
        t_now = time.time()
        if t_now - t0 < cache._min_dtime:
            # Log more information to troubleshoot
            lgr.error(f"Failing test with t0={t0}, t_now={t_now}, "
                      f"dt={t_now - t0}, min_dtime={cache._min_dtime}")
            for p in ("a.txt", "b.txt"):
                lgr.error(f"   {p}: {op.getmtime(path / p)}")
            raise  # if we were quick but still failed -- legit
    assert calls[-1] == [path, 0, None]

    # but if we sleep - should memoize
    time.sleep(cache._min_dtime * 1.1)
    check_new_memoread(1, 9)

    # and if we modify the file -- a new read
    time.sleep(cache._min_dtime * 1.1)
    (path / "c.txt").write_text("Gamma")
    ncalls = len(calls)
    assert memoread(path, 1) == 14
    assert len(calls) == ncalls + 1

    time.sleep(cache._min_dtime * 1.1)
    check_new_memoread(0, 14)

    # Check that symlinks should be dereferenced
    if not on_windows or (
        sys.version_info[:2] >= (3, 8) and not (on_windows and on_pypy)
    ):
        # realpath doesn't work right on Windows on pre-3.8 Python, and PyPy on
        # Windows doesn't support symlinks at all, so skip the test then.
        symlink1 = str(tmp_path / (fname + ".link1"))
        try:
            os.symlink(fname, symlink1)
        except OSError:
            pass
        if op.islink(symlink1):  # hopefully would just skip Windows if not supported
            ncalls = len(calls)
            assert memoread(symlink1, 0) == 14
            assert len(calls) == ncalls  # no new call

    # and if we "clear", would it still work?
    cache.clear()
    check_new_memoread(1, 14)


def test_memoize_path_persist(tmp_path):
    from subprocess import PIPE, run

    script = tmp_path / "script.py"
    cachedir = tmp_path / "cache"
    script.write_text(
        "from os.path import basename\n"
        "from fscacher import PersistentCache\n"
        f"cache = PersistentCache(path={str(cachedir)!r})\n"
        "\n"
        "@cache.memoize_path\n"
        "def func(path):\n"
        "    print('Running %s.' % basename(path), end='')\n"
        "    return 'DONE'\n"
        "\n"
        f"print(func({str(script)!r}))\n"
    )

    outputs = [
        run([sys.executable, str(script)], stdout=PIPE, stderr=PIPE) for i in range(3)
    ]
    print("Full outputs: %s" % repr(outputs))
    if b"File name too long" in outputs[0].stderr:
        # must be running during conda build which blows up paths with
        # _placehold_ers
        pytest.skip("seems to be running on conda and hitting the limits")
    assert outputs[0].stdout.strip().decode() == "Running script.py.DONE"
    for o in outputs[1:]:
        assert o.stdout.strip().decode() == "DONE"


def test_memoize_path_tokens(tmp_path, cache, cache_tokens):
    calls = []

    @cache.memoize_path
    def memoread(path, arg, kwarg=None):
        calls.append(["cache", path, arg, kwarg])
        with open(path) as f:
            return f.read()

    @cache_tokens.memoize_path
    def memoread_tokens(path, arg, kwarg=None):
        calls.append(["cache_tokens", path, arg, kwarg])
        with open(path) as f:
            return f.read()

    def check_new_memoread(call, arg, content, expect_first=True, expect_new=False):
        ncalls = len(calls)
        assert call(path, arg) == content
        assert len(calls) == ncalls + int(expect_first)
        assert call(path, arg) == content
        assert len(calls) == ncalls + int(expect_first) + int(expect_new)

    path = str(tmp_path / "file.dat")

    with open(path, "w") as f:
        f.write("content")

    time.sleep(cache._min_dtime * 1.1)
    # They both are independent, so both will cause a new readout
    check_new_memoread(memoread, 0, "content")
    check_new_memoread(memoread_tokens, 0, "content")


@pytest.mark.parametrize(
    "fscacher_value,mycache_value,cleared,ignored",
    [
        ("clear", None, True, False),
        ("q", "clear", True, False),
        ("ignore", "clear", True, False),
        ("clear", "", False, False),
        ("clear", "q", False, False),
        ("ignore", None, False, True),
        ("q", "ignore", False, True),
        ("clear", "ignore", False, True),
        ("ignore", "", False, False),
        ("ignore", "q", False, False),
    ],
)
def test_cache_control_envvar(
    mocker, monkeypatch, fscacher_value, mycache_value, cleared, ignored, tmp_path
):
    if fscacher_value is not None:
        monkeypatch.setenv("FSCACHER_CACHE", fscacher_value)
    else:
        monkeypatch.delenv("FSCACHER_CACHE", raising=False)
    if mycache_value is not None:
        monkeypatch.setenv("MYCACHE_CONTROL", mycache_value)
    else:
        monkeypatch.delenv("MYCACHE_CONTROL", raising=False)
    clear_spy = mocker.spy(PersistentCache, "clear")
    c = PersistentCache(path=tmp_path, envvar="MYCACHE_CONTROL")
    assert clear_spy.called is cleared
    assert c._ignore_cache is ignored


@pytest.mark.skipif(shutil.which("git-annex") is None, reason="git annex required")
def test_follow_moved_symlink(cache, tmp_path):
    calls = []

    @cache.memoize_path
    def memoread(path):
        calls.append([path])
        with open(path) as f:
            return f.read()

    def git(*args):
        subprocess.run(["git", *args], cwd=tmp_path, check=True)

    content = "This is test text.\n"
    git("init")
    git("annex", "init")
    (tmp_path / "file.txt").write_text(content)
    git("annex", "add", "file.txt")
    git("commit", "-m", "Create file")
    assert op.islink(tmp_path / "file.txt")

    assert memoread(tmp_path / "file.txt") == content
    assert len(calls) == 1
    assert memoread(tmp_path / "file.txt") == content
    assert len(calls) == 1

    git("mv", "file.txt", "text.txt")
    git("commit", "-m", "Rename file")

    assert memoread(tmp_path / "text.txt") == content
    assert len(calls) == 1
    assert memoread(tmp_path / "text.txt") == content
    assert len(calls) == 1

    (tmp_path / "subdir").mkdir()
    git("mv", "text.txt", op.join("subdir", "text.txt"))
    git("commit", "-m", "Move file")

    assert memoread(tmp_path / "subdir" / "text.txt") == content
    assert len(calls) == 1
    assert memoread(tmp_path / "subdir" / "text.txt") == content
    assert len(calls) == 1


def test_memoize_path_nonpath_arg(cache, tmp_path):
    calls = []

    @cache.memoize_path
    def memoread(filepath, arg, kwarg=None):
        calls.append([filepath, arg, kwarg])
        with open(filepath) as f:
            return f.read()

    path = str(tmp_path / "file.dat")
    with open(path, "w") as f:
        f.write("content")

    time.sleep(cache._min_dtime * 1.1)

    ncalls = len(calls)
    assert memoread(path, 1) == "content"
    assert len(calls) == ncalls + 1
    assert memoread(arg=1, filepath=path) == "content"
    assert len(calls) == ncalls + 1


def test_dir_fingerprint_order_irrelevant(tmp_path):
    start = time.time()
    file1 = tmp_path / "apple.txt"
    file1.write_text("Apple\n")
    os.utime(file1, (start - 1, start - 1))
    file2 = tmp_path / "banana.txt"
    file2.write_text("This is test text.\n")
    os.utime(file2, (start - 2, start - 2))
    file3 = tmp_path / "coconut.txt"
    file3.write_text("Lorem ipsum dolor sit amet, consectetur adipisicing elit\n")
    os.utime(file3, (start - 3, start - 3))
    df_tuples = []
    for file_list in [
        [file1, file2, file3],
        [file3, file2, file1],
        [file2, file1, file3],
    ]:
        dprint = DirFingerprint()
        for f in file_list:
            fprint = FileFingerprint.from_stat(os.stat(f))
            dprint.add_file(f, fprint)
        df_tuples.append(dprint.to_tuple())
    for i in range(1, len(df_tuples)):
        assert df_tuples[0] == df_tuples[i]


def test_memoize_non_pathlike_arg(cache, tmp_path):
    calls = []

    @cache.memoize_path
    def strify(x):
        calls.append(x)
        return str(x)

    path = tmp_path / "foo"
    path.touch()
    time.sleep(cache._min_dtime * 1.1)

    assert strify(path) == str(path)
    assert calls == [path]

    assert strify(42) == "42"
    assert calls == [path, 42]

    assert strify(path) == str(path)
    assert calls == [path, 42]

    assert strify(42) == "42"
    assert calls == [path, 42, 42]


@dataclass
class PathWrapper:
    path: Path

    def __fspath__(self) -> str:
        return str(self.path)

    def __str__(self) -> str:
        return str(self.path)


def test_memoize_pathlike_arg(cache, tmp_path):
    calls = []

    @cache.memoize_path
    def strify(x):
        calls.append(x)
        return str(x)

    path = tmp_path / "foo"
    path.touch()
    foo = PathWrapper(path)

    path2 = tmp_path / "bar"
    path2.touch()
    bar = PathWrapper(path2)

    time.sleep(cache._min_dtime * 1.1)

    assert strify(path) == str(path)
    assert calls == [path]

    assert strify(foo) == str(path)
    assert calls == [path]

    assert strify(bar) == str(tmp_path / "bar")
    assert calls == [path, bar]

    assert strify(path) == str(path)
    assert calls == [path, bar]

    assert strify(foo) == str(path)
    assert calls == [path, bar]

    assert strify(bar) == str(tmp_path / "bar")
    assert calls == [path, bar]


def test_memoize_path_exclude_kwargs(cache, tmp_path):
    calls = []

    @cache.memoize_path(exclude_kwargs=["extra"])
    def memoread_extra(path, arg, kwarg=None, extra=None):
        calls.append((path, arg, kwarg, extra))
        with open(path) as f:
            return f.read()

    path = tmp_path / "file.dat"
    path.write_text("content")

    time.sleep(cache._min_dtime * 1.1)

    assert memoread_extra(path, 1, extra="foo") == "content"
    assert calls == [(path, 1, None, "foo")]

    assert memoread_extra(path, 1, extra="bar") == "content"
    assert calls == [(path, 1, None, "foo")]

    assert memoread_extra(path, 1, kwarg="quux", extra="bar") == "content"
    assert calls == [(path, 1, None, "foo"), (path, 1, "quux", "bar")]

    path.write_text("different")

    time.sleep(cache._min_dtime * 1.1)

    assert memoread_extra(path, 1, extra="foo") == "different"
    assert calls == [
        (path, 1, None, "foo"),
        (path, 1, "quux", "bar"),
        (path, 1, None, "foo"),
    ]

    assert memoread_extra(path, 1, extra="bar") == "different"
    assert calls == [
        (path, 1, None, "foo"),
        (path, 1, "quux", "bar"),
        (path, 1, None, "foo"),
    ]



"""
Caching results of operations on heavy file trees

Visit <https://github.com/con/fscacher> for more information.
"""

from ._version import get_versions
from .cache import PersistentCache

__version__ = get_versions()["version"]
__author__ = "Center for Open Neuroscience"
__author_email__ = "debian@onerussian.com"
__license__ = "MIT"
__url__ = "https://github.com/con/fscacher"

__all__ = ["PersistentCache"]



import os.path
import sys
from setuptools import setup

# This is needed for versioneer to be importable when building with PEP 517.
# See <https://github.com/warner/python-versioneer/issues/193> and links
# therein for more information.
sys.path.append(os.path.dirname(__file__))

import versioneer

setup(
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
)
